import platform

import os


class Backend(object):

    def shortname(self):
        return type(self).__name__

    def start(self, func, *args, **kwargs):
        """Starts a tasklet/greenlet."""
        raise NotImplementedError()

    def run(self, func, *args, **kwargs):
        """Runs a tasklet up until it blocks or finishes."""
        raise NotImplementedError()

    def channel(self):
        """Returns a new channel."""
        raise NotImplementedError()

    def yield_(self):
        """Yields control for other tasklets/greenlets to run."""
        raise NotImplementedError()

    def resume(self, tasklet):
        """Runs the given tasklet/greenlet immediately."""
        raise NotImplementedError()

    def propagate_exc(self, errtype, *args):
        """Propagates an exception (created via ``errtype(*args)``)
        so the program hears it and it doesn't die lonely in a tasklet."""
        raise NotImplementedError()


# We can't easily use stackless on our CI,
# so don't worry about covering it.
def _make_stackless():  # pragma: no cover
    import stackless

    class StacklessBackend(Backend):
        def shortname(self):
            return 'stackless'

        def start(self, func, *args, **kwargs):
            return stackless.tasklet(func)(*args, **kwargs)

        def run(self, func, *args, **kwargs):
            t = self.start(func, *args, **kwargs)
            t.run()
            return t

        def channel(self):
            return stackless.channel()

        def yield_(self):
            return stackless.schedule()

        def resume(self, tasklet):
            tasklet.run()

        def propagate_exc(self, errtype, *args):
            stackless.getmain().throw(errtype, *args)

    return StacklessBackend()


def _make_gevent():
    import gevent
    import gevent.queue

    class Channel(gevent.queue.Channel):
        def send(self, value):
            self.put(value)

        def receive(self):
            return self.get()

    class GeventBackend(Backend):
        def shortname(self):
            return 'gevent'  # pragma: no cover

        def start(self, func, *args, **kwargs):
            greenlet = gevent.spawn(func, *args, **kwargs)
            return greenlet

        def run(self, func, *args, **kwargs):
            greenlet = self.start(func, *args, **kwargs)
            gevent.sleep()
            return greenlet

        def channel(self):
            return Channel()

        def yield_(self):
            gevent.sleep()

        def resume(self, tasklet):
            gevent.sleep()

        def propagate_exc(self, errtype, *args):
            raise errtype

    return GeventBackend()


NO_VALID_BACKEND_MSG = """
No valid backend could be created.
Valid backends are
gevent (for CPython, Stackless Python, or PyPy with newer version of gevent)
and
stackless (for Stackless Python or PyPy).
See goless.backends.calculate_backend for more details about what backend
is chosen under what conditions."""


class NoValidBackend(Exception):
    def __init__(self):
        Exception.__init__(self, NO_VALID_BACKEND_MSG)


class NullBackend(Backend):
    """Backend that raises a NoValidBackend when it is accessed or called.
    This allows goless to be imported, but not used,
    when no backend can be found.
    """

    def __getattribute__(self, item):
        raise NoValidBackend()

    def __call__(self, *args, **kwargs):
        raise NoValidBackend()


_default_backends = {
    'stackless': _make_stackless,
    'gevent': _make_gevent
}

is_pypy = platform.python_implementation() == 'PyPy'


def _calc_default(backends):
    try:
        if is_pypy and 'gevent' in backends:
            return backends['gevent']()
        if 'stackless' in backends:
            return backends['stackless']()
    except ImportError:
        pass
    raise SystemError('Swallow this, please.')


def calculate_backend(name_from_env, backends=None):
    """
    Calculates which backend to use with the following algorithm:

    - Try to read the GOLESS_BACKEND environment variable.
      Usually 'gevent' or 'stackless'.
      If a value is set but no backend is available or it fails to be created,
      this function will error.
    - Determine the default backend (gevent for PyPy, stackless for Python).
      If no default can be determined or created, continue.
    - Try to create all the runtimes and choose the first one to create
      successfully.
    - If no runtime can be created, return a NullBackend,
      which will error when accessed.
    """
    if backends is None:
        backends = _default_backends
    if name_from_env:
        if name_from_env not in backends:
            raise RuntimeError(
                'Invalid backend %r specified. Valid backends are: %s'
                % (name_from_env, _default_backends.keys()))
        return backends[name_from_env]()
    try:
        return _calc_default(backends)
    except SystemError:
        pass
    for maker in backends.values():
        # noinspection PyBroadException
        try:
            return maker()
        except Exception:
            pass
    return NullBackend()


current = calculate_backend(os.getenv('GOLESS_BACKEND', ''))



"""
goless provides Golang semantics built on top of
gevent, PyPy, or Stackless Python.
Use goless.chan to create a synchronous or buffered channel.
Use goless.go to start a goroutine.
Use goless.select like you would Go's reflect.Select function,
with goless.dcase, rcase, and scase.

goless is fully documented, tested, and has many examples.
See http://goless.readthedocs.org/ for more documentation.
"""

import logging
import sys

import traceback

from .backends import current as _be


# noinspection PyUnresolvedReferences
from .channels import chan, ChannelClosed
# noinspection PyUnresolvedReferences
from .selecting import dcase, rcase, scase, select


version_info = 0, 5, 0
version = '.'.join([str(v) for v in version_info])

__version__ = version
__author__ = 'Rob Galanakis'
__email__ = 'rob.galanakis@gmail.com'
__url__ = 'https://github.com/rgalanakis/goless'
__license__ = 'Apache'


def on_panic(etype, value, tb):
    """
    Called when there is an unhandled error in a goroutine.
    By default, logs and exits the process.
    """
    logging.critical(traceback.format_exception(etype, value, tb))
    _be.propagate_exc(SystemExit, 1)


def go(func, *args, **kwargs):
    """
    Run a function in a new tasklet, like a goroutine.
    If the goroutine raises an unhandled exception (*panics*),
    the :func:`goless.on_panic` will be called,
    which by default logs the error and exits the process.

    :param args: Positional arguments to ``func``.
    :param kwargs: Keyword arguments to ``func``.
    """
    def safe_wrapped(f):
        # noinspection PyBroadException
        try:
            f(*args, **kwargs)
        except:
            on_panic(*sys.exc_info())
    _be.start(safe_wrapped, func)



from .backends import current as _be


# noinspection PyPep8Naming,PyShadowingNames
class rcase(object):
    """
    A case that will ``chan.recv()`` when the channel is able to receive.
    """
    def __init__(self, chan):
        self.chan = chan

    def ready(self):
        return self.chan.recv_ready()

    def exec_(self):
        return self.chan.recv()


# noinspection PyPep8Naming,PyShadowingNames
class scase(object):
    """A case that will ``chan.send(value)`` when the channel is able to send."""
    def __init__(self, chan, value):
        self.chan = chan
        self.value = value

    def ready(self):
        return self.chan.send_ready()

    def exec_(self):
        self.chan.send(self.value)


# noinspection PyPep8Naming
class dcase(object):
    """The default case."""
    def ready(self):
        return False


def select(cases):
    """
    Select the first case that becomes ready.
    If a default case (:class:`goless.dcase`) is present,
    return that if no other cases are ready.
    If there is no default case and no case is ready,
    block until one becomes ready.

    See Go's ``reflect.Select`` method for an analog
    (http://golang.org/pkg/reflect/#Select).

    :param cases: List of case instances, such as
      :class:`goless.rcase`, :class:`goless.scase`, or :class:`goless.dcase`.
    :return: ``(chosen case, received value)``.
      If the chosen case is not an :class:`goless.rcase`, it will be None.
    """
    default = None
    for c in cases:
        if c.ready():
            return c, c.exec_()
        if isinstance(c, dcase):
            assert default is None, 'Only one default case is allowd.'
            default = c
    if default is not None:
        # noinspection PyCallingNonCallable
        return default, None

    while True:
        for c in cases:
            if c.ready():
                return c, c.exec_()
        _be.yield_()



from setuptools import setup

from goless import version, __author__, __email__, __url__, __license__

setup(
    name='goless',
    version=version,
    author=__author__,
    author_email=__email__,
    description="Provides a Go-like concurrent programming style in Python.",
    long_description="""goless provides Go-like channels and a select function
    in Python, written on top of gevent, PyPy, or Stackless Python.
    Using goless, you can write Go-style concurrent programs in Python.
    """,
    license=__license__,
    keywords='tasklet stackless go concurrent '
             'threading async gevent go golang',
    url=__url__,
    packages=['goless'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 3',
    ],
    test_suite='tests',
)



import collections as _collections
import sys

from .backends import current as _be


class ChannelClosed(Exception):
    """
    Exception raised to indicate a channel is closing or has closed.
    """


class GoChannel(object):
    """
    A **Go**-like channel that can be sent to, received from,
    and closed.
    Callers should never create this directly.
    Always use :func:`goless.chan` to create channels.
    """
    def __init__(self):
        self._closed = False

    def send(self, value=None):
        """
        Sends the value. Blocking behavior depends on the channel type.
        Unbufferred channels block if no receiver is waiting.
        Buffered channels block if the buffer is full.
        Asynchronous channels never block on send.

        If the channel is already closed,
        :class:`goless.ChannelClosed` will be raised.
        If the channel closes during a blocking ``send``,
        :class:`goless.ChannelClosed` will be raised. (#TODO)
        """
        if self._closed:
            raise ChannelClosed()
        self._send(value)

    def _send(self, value):
        raise NotImplementedError()

    def recv(self):
        """
        Receive a value from the channel.
        Receiving will always block if no value is available.
        If the channel is already closed, :class:`goless.ChannelClosed` will be raised.
        If the channel closes during a blocking ``recv``,
        :class:`goless.ChannelClosed` will be raised. (#TODO)
        """
        if self._closed and not self.recv_ready():
            raise ChannelClosed()
        got = self._recv()
        return got

    def _recv(self):
        raise NotImplementedError()

    def recv_ready(self):
        """
        Return True if there is a sender waiting,
        or there are items in the buffer.
        """
        raise NotImplementedError()

    def send_ready(self):
        """
        Return True if a receiver is waiting,
        or the buffer has room.
        """
        raise NotImplementedError()

    def close(self):
        """
        Closes the channel, not allowing further communication.
        Any blocking senders or receivers will be woken up and raise :class:`goless.ChannelClosed`.
        Receiving or sending to a closed channel will raise :class:`goless.ChannelClosed`.
        """
        self._closed = True

    def __iter__(self):
        return self

    def next(self):
        try:
            return self.recv()
        except ChannelClosed:
            raise StopIteration


class BufferedChannel(GoChannel):
    """
    BufferedChannel has several situations it must handle.

    When sending:

    1. If there is a receiver waiting send the value through the channel.
       A waiting receiver indicates the buffer was empty.
       It's uncertain in this case if there will be a context switch 
       (blocking the current task) or not.
    2. Else if there is no more room in the buffer,
       also send the value through the channel,
       blocking until a receiver for the value is available.
    3. Otherwise just append the value to the deque returning immediately.

    When receiving:

    1. If the buffer has items, pop and return the first value.
       Before returning, if there is a sender waiting,
       receive its value and append it to the buffer.
    2. Else just receive on the channel,
       blocking until a sender is available.
       Return the value from the sender.
    """

    def __init__(self, size):
        assert isinstance(size, int) and size >= 0
        GoChannel.__init__(self)
        self.maxsize = size
        self.values_deque = _collections.deque() if size else ()
        self.waiting_chan = _be.channel()

    def _send(self, value):
        buffer_size = len(self.values_deque)
        chan_balance = self.waiting_chan.balance
        assert buffer_size <= self.maxsize
        assert ((chan_balance < 0 and buffer_size == 0)
                or (chan_balance > 0 and buffer_size == self.maxsize)
                or chan_balance == 0)
        if chan_balance < 0 or buffer_size == self.maxsize:
            self.waiting_chan.send(value)
            if self._closed:
                raise ChannelClosed("Channel closed while sending")
        else:
            assert buffer_size < self.maxsize
            self.values_deque.append(value)

    def _recv(self):
        if self.values_deque:
            value = self.values_deque.popleft()
            if self.waiting_chan.balance > 0:
                self.values_deque.append(self.waiting_chan.receive())
        else:
            value = self.waiting_chan.receive()
            if self._closed:
                raise ChannelClosed("Channel closed while receiving")
        return value

    def recv_ready(self):
        return self.values_deque or self.waiting_chan.balance > 0

    def send_ready(self):
        return len(self.values_deque) < self.maxsize or self.waiting_chan.balance < 0

    def close(self):
        # This next yield gives a chance to all blocked receivers to return
        # before the channel is actually closed.
        # We think this is more "fair",
        # as these are blocking-but-ready tasklets that should have
        # the opportunity to finish.
        # Otherwise, for example, a sender could successfully send
        # but still get a ChannelClosed error.
        _be.yield_()
        # To make sure all pending tasklets are woken up,
        # we mark the channel closed and then spam out sends or receives if needed.
        # The tasklets will wake up, see the channel is closed,
        # and raise a ChannelClosed error.
        GoChannel.close(self)
        balance = self.waiting_chan.balance
        for _ in xrange(balance, 0):
            self.waiting_chan.send(None)
        for _ in xrange(balance):
            self.waiting_chan.receive()


class SyncChannel(BufferedChannel):
    """
    Channel that behaves synchronously.
    A recv blocks until a sender is available,
    and a sender blocks until a receiver is available.
    Implemented as a special case of BufferedChannel
    where the buffer size is 0.
    """

    def __init__(self):
        BufferedChannel.__init__(self, 0)


class AsyncChannel(BufferedChannel):
    """
    A channel where send never blocks,
    and recv blocks if there are no items in the buffer.
    Implemented as a special case of BufferedChannel
    where the buffer size is sys.maxint.
    """

    def __init__(self):
        BufferedChannel.__init__(self, sys.maxint)


def chan(size=0):
    """
    Returns a bidirectional channel.

    A 0 or None size indicates a blocking channel
    (the ``send`` method will block until a receiver is available,
    and the ``recv`` method will block until a sender is available).

    A positive integer value will return a channel with a buffer.
    Once the buffer is filled, the ``send`` method will block.
    When the buffer is empty, the ``recv`` method will block.

    A negative integer will return a channel that will
    never block when the ``send`` method is called.
    The ``recv`` method will block if the buffer is empty.

    :rtype: goless.channels.GoChannel
    """
    if not size:
        return SyncChannel()
    if size < 0:
        return AsyncChannel()
    return BufferedChannel(size)
